## Copyright (C) 2024-2025 Andreas Bertsatos <abertsatos@biol.uoa.gr>
##
## This file is part of the datatypes package for GNU Octave.
##
## This program is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free Software
## Foundation; either version 3 of the License, or (at your option) any later
## version.
##
## This program is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
## details.
##
## You should have received a copy of the GNU General Public License along with
## this program; if not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn  {datatypes} {} display (@var{C})
##
## Customized display of cell arrays.
##
## This function overloads Octave core @code{display} with respect to cell
## arrays and displays their contents in MATLAB like manner.
##
## @end deftypefn
function display (this)
  in_name = inputname (1);
  if (! isempty (in_name))
    fprintf ('%s =\n', in_name);
  endif
  __disp__ (this, in_name);
endfunction

